#ifndef __HS_LZSS_H
#define __HS_LZSS_H

//
// lzss.h
// (c)2002 Jonathan Bennett (jon@hiddensoft.com)
//
// STANDALONE CLASS
//
// This is a memory-to-memory implementation of the LZSS compression
// algorithm.  See the "lzss.txt" file for a hopefully useful explanation
// of how LZSS works.
//
// Note, this is a "as simple to follow as possible" piece of code, there are many
// more efficient and quickler (and cryptic) methods but this was done as a learning
// exercise.
//

//
// Speeds and compression levels
// Input   Winbits Lenbits  Compressed
// -----   ------- -------  ----------
// v0.11
// 2345984   13       4     1191746, 1m45s (greedy parsing)
//
// v0.20
// 2345984   13       4     1216648, 5s (32 limit hashchain, hashmem used=224316 bytes)
//
//
// The start of the compression stream will contain "other" items apart from data
// LZSS					4 bytes
// Uncompressed Size	4 bytes (1 ULONG)
// Window bits			1 byte
// Match bits			1 byte
// ...
// Compressed data
// ...
//


// Error codes
#define	HS_LZSS_E_OK			0				// OK
#define	HS_LZSS_E_BADCOMPRESS	1				// Compressed file would be bigger than source!
#define HS_LZSS_E_NOTLZSS		2				// Not a valid LZSS data stream

// Stream flags
#define HS_LZSS_LITERAL			0				// Just output the literal byte
#define HS_LZSS_MATCH			1				// Output a (offset, len) match pair

#define HS_LZSS_MINDATASIZE		32				// Arbitrary miniumum data size that we should attempt to compress

// Define our hash and hash table structure
#define LZSS_HASHTABLE_SIZE		4096			// Our hash function gives values 0-4095

#define LZSS_HASHCHAIN_MAX		32				// Maximum numbers of chains per hash value
												// Best value would be equal to our window size
												// but very slow.  You can get within 1% of this
												// best compression by using the value
												// winsize/256

#define LZSS_HASHNUMCHARS		3				// Number of chars we hash

typedef struct LZSS_Hash
{
	struct	LZSS_Hash *lpPrev;					// Previous entry in linked list
	ULONG	nPos;								// Position in data stream
	struct 	LZSS_Hash *lpNext;					// Next entry in linked list

} _LZSS_Hash;

												


class HS_LZSS
{
public:
	// Functions

	int		Compress(UCHAR *bData, UCHAR *bCompressedData, 
						ULONG nDataSize, ULONG *nCompressedSize,
						UINT nWindowBits, UINT nLenBits);

	int		Uncompress(UCHAR *bCompressedData, UCHAR *bData);

	int		GetUncompressedSize(UCHAR *bCompressedData, ULONG *nUncompressedSize);



private:
	// Variables
	ULONG	m_nDataStreamPos;					// Current position in the data stream
	ULONG	m_nCompressedStreamPos;				// Curent position in the compressed stream
	UCHAR	*m_bData;
	UCHAR	*m_bCompressedData;

	ULONG	m_nDataSize;						// The size of our uncompressed data
	ULONG	m_nCompressedSize;					// The size of our compressed data

	// Temporary variables used for the bit operations
	ULONG	m_nCompressedLong;					// Compressed stream temporary 32bit value
	int		m_nCompressedBitsUsed;				// Number of bits used in temporary value

	// Window size related variables
	UINT	m_nWindowBits;
	UINT	m_nWindowLen;
	UINT	m_nMatchLenBits;
	UINT	m_nMatchLen;
	UINT	m_nMinMatchLen;
	UINT	m_nWindowLenAdjust;
	UINT	m_nMatchLenAdjust;

	// Hash table
	struct	LZSS_Hash *m_HashTable[4096];		// Hash table (4096 key hash function)
	ULONG	m_nHashMemory, m_nHashMemoryMax;

	// Functions
	void	SetupWindowSize(UINT nWindowBits, UINT nLenBits);

	bool	FindMatches(UINT *nOffset, UINT *nLen);		// Searches for pattern matches
	void	FindMatches2(UINT *nOffset, UINT *nLen, UINT nHash, ULONG nAdjust);	// Searches for pattern matches

	// Bit operation functions
	void	CompressedStreamWriteBits(UINT nValue, UINT nNumBits);
	void	CompressedStreamReadBits(UINT *nValue, UINT nNumBits);

	// Hash table functions
	void	HashTableInit(void);
	void	HashTableFree(void);
	void	HashTableAdd(UINT nHash, ULONG nPos);
	void	HashTablePrune(UINT nHash, ULONG nOldestPos);
};


#endif
