//
// lzss_compress.cpp
// (c)2002 Jonathan Bennett (jon@hiddensoft.com)
//
//

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "..\include\lzss.h"


///////////////////////////////////////////////////////////////////////////////
// Compress()
///////////////////////////////////////////////////////////////////////////////

int HS_LZSS::Compress(UCHAR *bData, UCHAR *bCompressedData, 
						ULONG nDataSize, ULONG *nCompressedSize, 
						UINT nWindowBits, UINT nLenBits)
{
	UINT	nOffset, nLen;

	// Set up initial values
	m_nDataStreamPos		= 0;				// We are at the start of the input data
	m_nCompressedStreamPos	= 0;				// We are at the start of the compressed data
	m_nCompressedLong		= 0;				// Compressed stream temporary 32bit value
	m_nCompressedBitsUsed	= 0;				// Number of bits used in temporary value
	m_bCompressedData		= bCompressedData;	// Pointer to our output buffer
	m_bData					= bData;			// Pointer to our input buffer
	m_nDataSize				= nDataSize;		// Store the size of our input data
	m_nCompressedSize		= 0;				// No compressed data yet


	// If the input file is too small then there is a chance of 
	// buffer overrun, so just abort
	if (m_nDataSize < HS_LZSS_MINDATASIZE)
		return HS_LZSS_E_BADCOMPRESS;

	// Setup our user defined window size and length match size
	SetupWindowSize(nWindowBits, nLenBits);

	// Write out our file header info so the decompressor doesn't have to
	// guess about things like uncompressed buffer sizes
	// LZSS ID (4 bytes)
	CompressedStreamWriteBits('L', 8);	CompressedStreamWriteBits('Z', 8);
	CompressedStreamWriteBits('S', 8);	CompressedStreamWriteBits('S', 8);

	// Uncompressed size (4 bytes)
	CompressedStreamWriteBits( (m_nDataSize >> 16) & 0x0000ffff, 16);
	CompressedStreamWriteBits( m_nDataSize & 0x0000ffff, 16);

	// Window and match bits
	CompressedStreamWriteBits( m_nWindowBits, 8);
	CompressedStreamWriteBits( m_nMatchLenBits, 8);

	// Initialize our hash table
	HashTableInit();


	// Loop around until there is no more data
	while (m_nDataStreamPos < m_nDataSize)
	{
		// If at any point, we reach within 16 bytes of the end of our output buffer, abort!
		// See the end of this while loop for details of what is output after we have
		// finished compressing
		if ( m_nCompressedSize >= (m_nDataSize - 16) )
			return HS_LZSS_E_BADCOMPRESS;

		// Search for matches
		if ( FindMatches(&nOffset, &nLen) == true )
		{
			// A match was found, write it out
			CompressedStreamWriteBits(HS_LZSS_MATCH, 1);
			CompressedStreamWriteBits(nOffset - m_nWindowLenAdjust, m_nWindowBits);
			CompressedStreamWriteBits(nLen - m_nMatchLenAdjust, m_nMatchLenBits);
		} 
		else
		{
			// No matches were found, just write out the literal byte
			CompressedStreamWriteBits(HS_LZSS_LITERAL, 1);
			CompressedStreamWriteBits(m_bData[m_nDataStreamPos++], 8);

		} // End If

	} // End while

	// Write out 2 bytes (16 bits) of nothing to help avoid overruns when decoding
	// Lazy....I know.
	CompressedStreamWriteBits(0, 16);

	// Flush the compressed bitstream (pad with zeros if required)
	CompressedStreamWriteBits(0, 0);

	// Free up our hash table
	HashTableFree();

	// Return the compressed data size
	*nCompressedSize = m_nCompressedSize;

	return HS_LZSS_E_OK;						// Return with success message

} // Compress()


///////////////////////////////////////////////////////////////////////////////
// SetupWindowSize()
//
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::SetupWindowSize(UINT nWindowBits, UINT nLenBits)
{
	UINT nTemp;

	m_nWindowBits			= nWindowBits;
	m_nMatchLenBits			= nLenBits;
	m_nMatchLenAdjust		= 0;

	// Setup initial window and bit lens (may be adjusted later for efficiency)
	m_nWindowLen	= ((UINT)pow(2, (double)m_nWindowBits)) - 1;
	m_nMatchLen		= ((UINT)pow(2, (double)m_nMatchLenBits)) - 1;

	// How many bits does a match offset take up?
	nTemp = m_nWindowBits + m_nMatchLenBits;

	// Each literal byte takes up 9 bits when encoded, what is the "break even"?
	// Force the break even to be at least 3 (hash table uses 3 chars)
	m_nMinMatchLen = (nTemp / 9) + 1;
	if (m_nMinMatchLen < LZSS_HASHNUMCHARS)
		m_nMinMatchLen = LZSS_HASHNUMCHARS;

	// Based on this information we adjust our lengths
	m_nWindowLenAdjust	= m_nMinMatchLen - 1;
	m_nWindowLen		= m_nWindowLen + m_nWindowLenAdjust;
	m_nMatchLenAdjust	= m_nMinMatchLen - 1;
	m_nMatchLen			= m_nMatchLen + m_nMatchLenAdjust;

	printf("Number of window bits: %d (0-%d)\n", m_nWindowBits, m_nWindowLen);
	printf("Number of length bits: %d (0-%d)\n", m_nMatchLenBits, m_nMatchLen);
	printf("Minimum match len    : %d\n", m_nMinMatchLen);
	printf("Window len adjust    : %d\n", m_nWindowLenAdjust);
	printf("Match len adjust     : %d\n", m_nMatchLenAdjust);

} // SetupWindowSize()


///////////////////////////////////////////////////////////////////////////////
// CompressedStreamWriteBits()
//
// Will write a number of bits (variable) into the compressed data stream
//
// When there are no more bits to send, you should call the function with the
// parameters 0, 0 to make sure that any left over bits are flushed into the
// compressed stream
//
// Note 16 bits is the maximum allowed value for this function!!!!!
// ================================================================
// This equates to a maximum window size of 65535
//
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::CompressedStreamWriteBits(UINT nValue, UINT nNumBits)
{
	if (nNumBits == 0)
	{
		// We have been asked to finish, flush remaining by using up remaining
		// unused bits with zeros

		// Bits remaining = 32 minus bits used
		m_nCompressedLong = m_nCompressedLong << (32 - m_nCompressedBitsUsed);

		// Now write out all 32 bits
		m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 24);
		m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 16);
		m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 8);
		m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong);

		// We have just stored 4 bytes of data, increase our size count
		m_nCompressedSize = m_nCompressedSize + 4;

		// All done
		m_nCompressedBitsUsed = 0;

		return;
	}

	while (nNumBits > 0)
	{
		nNumBits--;

		// Make room for another bit (shift left once)
		m_nCompressedLong = m_nCompressedLong << 1;

		// Merge (OR) our value into the temporary long
		m_nCompressedLong = m_nCompressedLong | (long)((nValue >> nNumBits) & 0x00000001);
		//printf("%d ", ((nValue >> nNumBits) & 0x00000001));

		// Update how many bits we are using (add 1)
		m_nCompressedBitsUsed++;

		// Now check if we have filled our temporary long with bits (32bits)
		if (m_nCompressedBitsUsed == 32)
		{
			// We now need to dump the highest 16 bits into our compressed
			// stream.  Highest order 8 bits first
			m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 24);
			m_bCompressedData[m_nCompressedStreamPos++] = (UCHAR)(m_nCompressedLong >> 16);

			// We have just stored 2 bytes of data, increase our size count
			m_nCompressedSize = m_nCompressedSize + 2;

			// We've now written out 16 bits so make more room (16 bits more room :) )
			m_nCompressedBitsUsed = m_nCompressedBitsUsed - 16;
		}

	} // End while


} // CompressedStreamWriteBits()


///////////////////////////////////////////////////////////////////////////////
// FindMatches()
///////////////////////////////////////////////////////////////////////////////

bool HS_LZSS::FindMatches(UINT *nOffset, UINT *nLen)
{
	UINT	nHash;
	UINT	i;
	ULONG	nWPos;

	// Generate a hash of the next three chars
	nHash =((40543*((((m_bData[m_nDataStreamPos]<<4)^m_bData[m_nDataStreamPos+1])<<4)^m_bData[m_nDataStreamPos+2]))>>4) & 0xFFF;

	// Search for matches
	FindMatches2(nOffset, nLen, nHash, 0);

	// Whether we matched our not, add the current hash to the table for future searching
	// Get our window start position, if the window would take us beyond
	// the start of the file, just use 0
	if (m_nDataStreamPos < m_nWindowLen)
		nWPos = 0;
	else
		nWPos = m_nDataStreamPos - m_nWindowLen;

	HashTablePrune(nHash, nWPos);			// Remove old entries for this hash
	HashTableAdd(nHash, m_nDataStreamPos);	// Made a note of the hash at this data position


	// If we matched, then return the best values
	if ( (*nOffset == 0) )						// No good matches
		return false;							// No match
	else
	{
		// We made a match of length nLen, we are about to skip forward this far in the
		// data stream, but lets not forget to hash all the values inbetween or we will
		// miss some matches on future searches of this part of the data

		m_nDataStreamPos++;							// Already added 1 hash to table, move on
		nWPos++;

		for (i=1; i<(*nLen); i++)
		{
			nHash =((40543*((((m_bData[m_nDataStreamPos]<<4)^m_bData[m_nDataStreamPos+1])<<4)^m_bData[m_nDataStreamPos+2]))>>4) & 0xFFF;
			HashTablePrune(nHash, nWPos);
			HashTableAdd(nHash, m_nDataStreamPos);	// Made a note of the hash at this data position
			m_nDataStreamPos++;
			nWPos++;
		}

		return true;							// Match! :)
	}


} // FindMatches()


///////////////////////////////////////////////////////////////////////////////
// FindMatches2()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::FindMatches2(UINT *nOffset, UINT *nLen, UINT nHash, ULONG nAdjust)
{
	ULONG	nTempWPos, nWPos, nDPos;	// Temp Window and Data position markers
	ULONG	nMaxWPos, nDataStreamPos, nDataSize;	// Pure temp values for speeding up loops
	ULONG	nTempLen;					// Temp vars 
	ULONG	nBestOffset, nBestLen;		// Stores the best match so far
	UCHAR	*bData;
	bool	bLoop;
	struct LZSS_Hash *lpTempHash;

	// Reset all variables
	nBestOffset = 0;
	nBestLen	= m_nMinMatchLen - 1;

	// Get our window start position, if the window would take us beyond
	// the start of the file, just use 0
	if (m_nDataStreamPos < m_nWindowLen)
		nWPos = 0;
	else
		nWPos = m_nDataStreamPos - m_nWindowLen;

	// Before we start searching the hash table, remove old values for this hash
	// based on our window position
	HashTablePrune(nHash, nWPos);

	// Store some temp values to speed up the loop (a little)
	nDataStreamPos	= m_nDataStreamPos;
	nMaxWPos		= nDataStreamPos - m_nMinMatchLen;
	bData			= m_bData;
	nDataSize		= m_nDataSize;
	

	// Main loop
	lpTempHash = m_HashTable[nHash];			// Get our match from the hash table
	bLoop = true;

	while ( (lpTempHash != NULL) && (bLoop == true))
	{
		nTempWPos	= lpTempHash->nPos;			// The position of the supposed match
		nDPos		= nDataStreamPos;
		nTempLen	= 0;
		while ( (bData[nTempWPos] == bData[nDPos]) && (nTempWPos < nDataStreamPos) &&
				nDPos < nDataSize) 
		{
			nTempLen++;
			nTempWPos++;
			nDPos++;
		} // End while

		// See if this match was better than previous match
		if (nTempLen > nBestLen)
		{
			if (nTempLen >= m_nMatchLen)
			{
				nBestLen	= m_nMatchLen;
				nBestOffset = nDataStreamPos - lpTempHash->nPos;
				bLoop = false;					// Force the while loop to end
			}
			else
			{
				nBestLen	= nTempLen;
				nBestOffset = nDataStreamPos - lpTempHash->nPos;
			}
		}

		lpTempHash = lpTempHash->lpNext;

	} // End while

	// Setup our return values of bestoffset and bestlen, bestoffset will be 0
	// if no good matches were match
	*nOffset	= nBestOffset;
	*nLen		= nBestLen;

} // FindMatches()


///////////////////////////////////////////////////////////////////////////////
// HashTableInit()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::HashTableInit(void)
{
	int i;

	for (i=0; i<LZSS_HASHTABLE_SIZE; i++)
		m_HashTable[i] = NULL;

	m_nHashMemory = m_nHashMemoryMax = 0;
}


///////////////////////////////////////////////////////////////////////////////
// HashTableFree()
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::HashTableFree(void)
{
	int i;
	struct LZSS_Hash *lpTempNext, *lpTempHash;

	for (i=0; i<LZSS_HASHTABLE_SIZE; i++)
	{
		lpTempHash = m_HashTable[i];
		while (lpTempHash != NULL)
		{
			// Go through the chain freeing each entry
			lpTempNext = lpTempHash->lpNext;
			free(lpTempHash);
			lpTempHash = lpTempNext;
		}
	}

	printf("\nHash memory = %d\n", m_nHashMemoryMax);

} // HashTableFree()


///////////////////////////////////////////////////////////////////////////////
// HashTableAdd()
// Adds a hash entry of "nPos" at index "nHash"
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::HashTableAdd(UINT nHash, ULONG nPos)
{
	int i;
	struct LZSS_Hash *lpTempNext, *lpTempHash;
	lpTempHash = m_HashTable[nHash];

	if (lpTempHash == NULL)
	{
		// Initial entry
		lpTempHash = (_LZSS_Hash*)malloc(sizeof(_LZSS_Hash));
		m_nHashMemory += sizeof(_LZSS_Hash);
		lpTempHash->lpPrev	= NULL;
		lpTempHash->lpNext	= NULL;
		lpTempHash->nPos	= nPos;
		m_HashTable[nHash] = lpTempHash;
		return;
	}

	// Existing entry, traverse the chain until we find an empty entry or 
	// exceed the size limit
	i = 0;
	while (i < LZSS_HASHCHAIN_MAX)
	{
		if (lpTempHash->lpNext == NULL)
		{
			// Found an empty entry
			lpTempNext = (_LZSS_Hash*)malloc(sizeof(_LZSS_Hash));
			m_nHashMemory += sizeof(_LZSS_Hash);
			lpTempHash->lpNext	= lpTempNext;
			lpTempNext->lpPrev	= lpTempHash;
			lpTempNext->lpNext	= NULL;
			lpTempNext->nPos	= nPos;
			i = LZSS_HASHCHAIN_MAX;				// No more searching required
		}

		lpTempHash = lpTempHash->lpNext;
		i++;
	} // End while

	//if (i == LZSS_HASHCHAIN_MAX)
	//	printf("m");

} // HashTableAdd()


///////////////////////////////////////////////////////////////////////////////
// HashTablePrune()
// Removes any hash entries older than "nOldestPos"
///////////////////////////////////////////////////////////////////////////////

void HS_LZSS::HashTablePrune(UINT nHash, ULONG nOldestPos)
{
	struct LZSS_Hash *lpTempNext, *lpTempHash, *lpTempPrev;

	lpTempHash = m_HashTable[nHash];

	if (m_nHashMemory > m_nHashMemoryMax)
		m_nHashMemoryMax = m_nHashMemory;

	// Traverse the chain (if there is one) and remove entries that are too
	// old to be used in a match
	
	while (lpTempHash != NULL)
	{
		lpTempNext	= lpTempHash->lpNext;

		if ((lpTempHash->nPos) < nOldestPos)
		{
			// Slip this entry out of the list
			lpTempPrev	= lpTempHash->lpPrev;		
			
			if (lpTempPrev != NULL)
				lpTempPrev->lpNext	= lpTempNext;
			else
				m_HashTable[nHash] = lpTempNext;	// New initial entry in table

			if (lpTempNext != NULL)
				lpTempNext->lpPrev	= lpTempHash->lpPrev;

			free(lpTempHash);
			m_nHashMemory -= sizeof(_LZSS_Hash);
		}

		lpTempHash = lpTempNext;

	} // End while

} // HashTablePrune()
